@@ -1,11 +1,10 @@
-
 /***************************************************************************
  * nsock_core.c -- This contains the core engine routines for the nsock    *
  * parallel socket event library.                                          *
  *                                                                         *
  ***********************IMPORTANT NSOCK LICENSE TERMS***********************
  *                                                                         *
- * The nsock parallel socket event library is (C) 1999-2003 Insecure.Com   *
+ * The nsock parallel socket event library is (C) 1999-2004 Insecure.Com   *
  * LLC This library is free software; you may redistribute and/or          *
  * modify it under the terms of the GNU General Public License as          *
  * published by the Free Software Foundation; Version 2.  This guarantees  *
@@ -14,6 +13,15 @@
  * may be willing to sell alternative licenses (contact                    *
  * sales@insecure.com ).                                                   *
  *                                                                         *
+ * As a special exception to the GPL terms, Insecure.Com LLC grants        *
+ * permission to link the code of this program with any version of the     *
+ * OpenSSL library which is distributed under a license identical to that  *
+ * listed in the included Copying.OpenSSL file, and distribute linked      *
+ * combinations including the two. You must obey the GNU GPL in all        *
+ * respects for all of the code used other than OpenSSL.  If you modify    *
+ * this file, you may extend this exception to your version of the file,   *
+ * but you are not obligated to do so.                                     *
+ *                                                                         * 
  * If you received these files with a written license agreement stating    *
  * terms other than the (GPL) terms above, then that alternative license   *
  * agreement takes precedence over this comment.                          *
@@ -45,7 +53,7 @@
  *                                                                         *
  ***************************************************************************/
 
-/* $Id: nsock_core.c,v 1.16 2003/09/16 06:04:55 fyodor Exp $ */
+/* $Id: nsock_core.c 2535 2005-01-31 20:40:45Z fyodor $ */
 
 #include "nsock_internal.h"
 #include "gh_list.h"
@@ -86,6 +94,8 @@ static int wait_for_events(mspool *ms, int msec_timeout) {
 
   int event_msecs; /* Msecs before an event goes off */
   int combined_msecs;
+  int sock_err = 0;
+  int socketno = 0;
   struct timeval select_tv;
   struct timeval *select_tv_p;
 
@@ -121,19 +131,40 @@ static int wait_for_events(mspool *ms, int msec_timeout) {
       select_tv_p = NULL;
     }
     
-    
-    /* Set up the descriptors for select */
-    ms->mioi.fds_results_r = ms->mioi.fds_master_r;
-    ms->mioi.fds_results_w = ms->mioi.fds_master_w;
-    ms->mioi.fds_results_x = ms->mioi.fds_master_x;
-    
-    ms->mioi.results_left = select(ms->mioi.max_sd + 1, &ms->mioi.fds_results_r, &ms->mioi.fds_results_w, &ms->mioi.fds_results_x, select_tv_p);
-    gettimeofday(&nsock_tod, NULL); /* Who knows how long select sat around for */
+    /* Figure out whether there are any FDs in the sets, as @$@!$#
+       Windows returns WSAINVAL (10022) if you call a select() with no
+       FDs, even though the Linux man page says that doing so is a
+       good, reasonably portable way to sleep with subsecond
+       precision.  Sigh */
+    for(socketno = ms->mioi.max_sd; socketno >= 0; socketno--) {
+      if(FD_ISSET(socketno, &ms->mioi.fds_master_r) ||
+	 FD_ISSET(socketno, &ms->mioi.fds_master_w) ||
+	 FD_ISSET(socketno, &ms->mioi.fds_master_x))
+	break;
+      else ms->mioi.max_sd--;
+    }
 
-  } while (ms->mioi.results_left == -1 && socket_errno() == EINTR);
+    if (ms->mioi.max_sd < 0) {
+      ms->mioi.results_left = 0;
+      if (combined_msecs > 0)
+	usleep(combined_msecs * 1000);
+    } else {
+
+      /* Set up the descriptors for select */
+      ms->mioi.fds_results_r = ms->mioi.fds_master_r;
+      ms->mioi.fds_results_w = ms->mioi.fds_master_w;
+      ms->mioi.fds_results_x = ms->mioi.fds_master_x;
+      
+      ms->mioi.results_left = select(ms->mioi.max_sd + 1, &ms->mioi.fds_results_r, &ms->mioi.fds_results_w, &ms->mioi.fds_results_x, select_tv_p);
+      if (ms->mioi.results_left == -1)
+	sock_err = socket_errno();
+    }
+
+    gettimeofday(&nsock_tod, NULL); /* Due to usleep or select delay */
+  } while (ms->mioi.results_left == -1 && sock_err == EINTR);
   
-  if (ms->mioi.results_left == -1 && socket_errno() != EINTR) {
-    ms->errnum = socket_errno();
+  if (ms->mioi.results_left == -1 && sock_err != EINTR) {
+    ms->errnum = sock_err;
     return -1;
   }
  
@@ -288,23 +319,17 @@ void handle_connect_result(mspool *ms, msevent *nse,
 #if HAVE_OPENSSL
   if (nse->type == NSE_TYPE_CONNECT_SSL && !nse->event_done) {
     /* Lets now start/continue/finish the connect! */
-    if (iod->ssl_session)
-      SSL_set_session(iod->ssl, iod->ssl_session);
+    if (iod->ssl_session) {    
+     rc = SSL_set_session(iod->ssl, iod->ssl_session);
+     if (rc == 0) { printf("Uh-oh: SSL_set_session() failed - please tell Fyodor\n"); }
+     iod->ssl_session = NULL; /* No need for this any more */
+    }
     rc = SSL_connect(iod->ssl);
     /* printf("DBG: SSL_connect()=%d", rc); */
     if (rc == 1) {
       /* Woop!  Connect is done! */
       nse->event_done = 1;
       nse->status = NSE_STATUS_SUCCESS;
-
-      if (iod->ssl_session == SSL_get0_session(iod->ssl)) {
-        // nothing to do - old is same as current.
-      } else {
-        // Free old session, load current one.
-	if (iod->ssl_session)
-	  SSL_SESSION_free(iod->ssl_session);
-	iod->ssl_session = SSL_get1_session(iod->ssl);
-      }
     } else {
       sslerr = SSL_get_error(iod->ssl, rc);
       if (rc == -1 && sslerr == SSL_ERROR_WANT_READ) {
@@ -422,7 +447,7 @@ static int do_actual_read(mspool *ms, msevent *nse) {
   char buf[8192];
   int buflen = 0;
   msiod *iod = nse->iod;
-  int err;
+  int err = 0;
   int max_chunk = NSOCK_READ_CHUNK_SIZE;
   int startlen = FILESPACE_LENGTH(&nse->iobuf);
 
@@ -431,30 +456,32 @@ static int do_actual_read(mspool *ms, msevent *nse) {
 
   if (!iod->ssl) {
     /* Traditional read() - no SSL - using recv() because that works better on Windows */
-    while(((buflen = recv(iod->sd, buf, sizeof(buf), 0)) > 0) ||
-	  ( buflen == -1 && socket_errno() == EINTR)) {
-      if (fscat(&nse->iobuf, buf, buflen) == -1) {
-	nse->event_done = 1;
-	nse->status = NSE_STATUS_ERROR;
-	nse->errnum = ENOMEM;
-	return -1;
+    do {
+      buflen = recv(iod->sd, buf, sizeof(buf), 0);
+      if (buflen == -1) err = socket_errno();
+      if (buflen > 0) {
+	if (fscat(&nse->iobuf, buf, buflen) == -1) {
+	  nse->event_done = 1;
+	  nse->status = NSE_STATUS_ERROR;
+	  nse->errnum = ENOMEM;
+	  return -1;
+	}
+
+	// Sometimes a service just spews and spews data.  So we return
+	// after a somewhat large amount to avoid monopolizing resources
+	// and avoid DOS attacks.
+	if (FILESPACE_LENGTH(&nse->iobuf) > max_chunk)
+	  return FILESPACE_LENGTH(&nse->iobuf) - startlen;
+	
+	// No good reason to read again if we we were successful in the read but
+	// didn't fill up the buffer.  I'll insist on it being TCP too, because I
+	// think UDP might only give me one packet worth at a time (I dunno ...).
+	if (buflen > 0 && buflen < sizeof(buf) && iod->lastproto == IPPROTO_TCP)
+	  return FILESPACE_LENGTH(&nse->iobuf) - startlen;
       }
-      
-      // Sometimes a service just spews and spews data.  So we return
-      // after a somewhat large amount to avoid monopolizing resources
-      // and avoid DOS attacks.
-      if (FILESPACE_LENGTH(&nse->iobuf) > max_chunk)
-	return FILESPACE_LENGTH(&nse->iobuf) - startlen;
-      
-      // No good reason to read again if we we were successful in the read but
-      // didn't fill up the buffer.  I'll insist on it being TCP too, because I
-      // think UDP might only give me one packet worth at a time (I dunno ...).
-      if (buflen > 0 && buflen < sizeof(buf) && iod->lastproto == IPPROTO_TCP)
-	return FILESPACE_LENGTH(&nse->iobuf) - startlen;
-    }
+    } while (buflen > 0 || (buflen == -1 && err == EINTR));
 
     if (buflen == -1) {
-      err = socket_errno();
       if (err != EINTR && err != EAGAIN) {
 	nse->event_done = 1;
 	nse->status = NSE_STATUS_ERROR;
@@ -924,7 +951,7 @@ void nsock_trace_handler_callback(mspool *ms, msevent *nse) {
   nsi = nse->iod;
 
   if (nse->status == NSE_STATUS_ERROR) {
-    snprintf(errstr, sizeof(errstr), "[%s] ", strerror(nse->errnum));
+    snprintf(errstr, sizeof(errstr), "[%s (%d)] ", strerror(nse->errnum), nse->errnum);
   } else errstr[0] = '\0';
 
   // Some types have special tracing treatment
